package Flat;

import java.util.*;

class myStm {
    Tree.Stm stm;
    myStm pred, succ;

    myStm(Tree.Stm s) {
        stm=s; pred=null; succ=null;
    }

    public void insertStm(myStm s) {
        s.pred=this;
	s.succ=null;

	succ=s;
    }

    Tree.Stm toStm() {
        return stm;
    }	
}

class myStmList {
    myStm first, last;

    myStmList() {
      first=null;
      last=null;
    }

    /* insert a statement at teh tail of myStmList */
    void add(Tree.Stm stm) {
        myStm s=new myStm(stm);

        if (first==null) {
	    first=s;
	    last=s;
	} else {
	    last.insertStm(s);
	    last=s;
	}
    }

    Tree.StmList toStmList() {
        if (last==null) 
	    return null;
	else {
	    Tree.StmList list=new Tree.StmList(last.toStm(), null);

	    myStm curstm=last.pred;
	    while (curstm!=null) {
	        list=new Tree.StmList(curstm.toStm(), list);
		curstm=curstm.pred;
	    }

	    return list;
	}
    }
}

public class Flat {

  static boolean isRealAtom(Tree.Exp exp) {
      if (exp instanceof Tree.CONST) return true;
      else if (exp instanceof Tree.TEMP) return true;
      else if (exp instanceof Tree.NAME) return true;
      else return false;
  }

  static boolean isAtomExp(Tree.Exp exp) {
      if (exp instanceof Tree.CONST) return true;
      else if (exp instanceof Tree.TEMP) return true;
      else if (exp instanceof Tree.NAME) return true;
      else if (exp instanceof Tree.MEM) {
	    Tree.MEM mem=(Tree.MEM)exp;
          Tree.Exp memExp=mem.exp;

          if (memExp instanceof Tree.CONST) return true;
	  if (memExp instanceof Tree.TEMP) return true;
          if (memExp instanceof Tree.BINOP) {
	        Tree.BINOP bop=(Tree.BINOP)memExp;

		  if (bop.binop==Tree.BINOP.PLUS || bop.binop==Tree.BINOP.MINUS) {
		      Tree.Exp subexp;

		      subexp=bop.left;
		      if ((subexp instanceof Tree.TEMP)
		          || (subexp instanceof Tree.CONST)) {
			  subexp=bop.right;
		          if ((subexp instanceof Tree.TEMP)
		              || (subexp instanceof Tree.CONST)) {
			      return true;
			  } else
			      return false;
		      }
		      else
		          return false;
		  } else
		      return false;
	      } else
	          return false;  
      }
      else if (exp instanceof Tree.SIMPLEVAR) {
          Tree.SIMPLEVAR sv=(Tree.SIMPLEVAR)exp;

          Tree.Exp lowExp=sv.lowExp;

          if (lowExp instanceof Tree.MEM) {
		  return isAtomExp(lowExp);
  	    } else {
	  //    throw new FlatException("Not a MEM in SIMPLEVAR");
	      return true;
          } 
      }
      else return false;
  }
  static Tree.Exp simplifyExp(Tree.Exp exp, myStmList list) {
    if (isAtomExp(exp)) {
        return exp;
    }

    if (exp instanceof Tree.BINOP) {
        Tree.BINOP bop=(Tree.BINOP)exp;
        Tree.Exp left, right;
        Tree.Exp re;

	left=simplifyExp(bop.left, list);
	right=simplifyExp(bop.right, list);
        re=new Tree.TEMP(new Temp.Temp()); 
	list.add(new Tree.MOVE(re, new Tree.BINOP(bop.binop, left, right)));
	return re;
        	
    } 
    else if (exp instanceof Tree.CALL) { 
        Tree.CALL call=(Tree.CALL)exp;

        // build argument vector
	Vector args=new Vector();

        for (Tree.ExpList arg=call.args; arg!=null; arg=arg.tail) {
	    if (arg.head==null) {
	        throw new FlatException("arg.head is null!");
	    }
            args.addElement(simplifyExp(arg.head, list));
        }

        // construct new call
        if (call.args==null) {
            return exp;
        } else {
	    Tree.ExpList argsList=new Tree.ExpList((Tree.Exp)args.lastElement(), null);
  	    for (int i=args.size()-2; i>=0; i--) {
	        argsList=new Tree.ExpList((Tree.Exp)(args.elementAt(i)), argsList);
	    }

            Tree.TEMP re=new Tree.TEMP(new Temp.Temp());

            list.add(new Tree.MOVE(re, new Tree.CALL(call.func, argsList)));

	    return re;
        }
    }
    else if (exp instanceof Tree.MEM) {
        Tree.MEM mem=(Tree.MEM)exp;
        Tree.Exp s;

        s=simplifyExp(mem.exp, list);
        Tree.MEM newMem=new Tree.MEM(s);

//        Tree.TEMP re=new Tree.TEMP(new Temp.Temp());
//        list.add(new Tree.MOVE(re, newMem));

        return (Tree.Exp)newMem;
    }
    else if (exp instanceof Tree.SIMPLEVAR) {
        // not an atom simplevar
        Tree.SIMPLEVAR sv=(Tree.SIMPLEVAR)exp;
        Tree.Exp s;

        s=simplifyExp(sv.lowExp, list);
        Tree.SIMPLEVAR re=new Tree.SIMPLEVAR(sv.name, s, sv.acc, sv.isLocal);

  	  return (Tree.Exp)re;
    }
    else if (exp instanceof Tree.FIELDVAR) {
        // evaluate exps
	Tree.FIELDVAR fv=(Tree.FIELDVAR)exp;
	Tree.Exp fRecord, fOffset;

	fRecord=simplifyExp(fv.record, list);
	fOffset=simplifyExp(fv.offset, list);
	if (!isRealAtom(fRecord)) {
	    Tree.TEMP t=new Tree.TEMP(new Temp.Temp());
	    list.add(new Tree.MOVE(t, fRecord));
	    fRecord=t;
	}
	if (!isRealAtom(fOffset)) {
	    Tree.TEMP t=new Tree.TEMP(new Temp.Temp());
	    list.add(new Tree.MOVE(t, fOffset));
	    fOffset=t;
	}

/*	fRecord=reduceToAtom(simplifyExp(fv.record, list), list);
	fOffset=reduceToAtom(simplifyExp(fv.offset, list), list);

      Tree.FIELDVAR newRv=new Tree.FIELDVAR(fRecord, fv.field, fOffset);
	Tree.Exp re=new Tree.TEMP(new Temp.Temp());
	list.add(new Tree.MOVE(re, newRv));
*/
      Tree.FIELDVAR newFv=new Tree.FIELDVAR(fRecord, fv.field, fOffset);
      
//      Tree.TEMP re=new Tree.TEMP(new Temp.Temp());
//      list.add(new Tree.MOVE(re, newFv));

      return (Tree.Exp)newFv;
    }
    else if (exp instanceof Tree.SUBSCRIPTVAR) {
        // evaluate exp
        Tree.SUBSCRIPTVAR sv=(Tree.SUBSCRIPTVAR)exp;
	Tree.Exp sArray, sIndex, sWs;

	sArray=simplifyExp(sv.array, list);
	sIndex=simplifyExp(sv.index, list);
	sWs=simplifyExp(sv.ws, list);

	if (!isRealAtom(sArray)) {
	    Tree.TEMP t=new Tree.TEMP(new Temp.Temp());
	    list.add(new Tree.MOVE(t, sArray));
	    sArray=t;
	}
	if (!isRealAtom(sIndex)) {
	    Tree.TEMP t=new Tree.TEMP(new Temp.Temp());
	    list.add(new Tree.MOVE(t, sIndex));
	    sIndex=t;
	}
/*	sArray=reduceToAtom(simplifyExp(sv.array, list), list);
        sIndex=reduceToAtom(simplifyExp(sv.index, list), list);

	Tree.SUBSCRIPTVAR newSv=new Tree.SUBSCRIPTVAR(sArray, sIndex, sWs);
	Tree.Exp re=new Tree.TEMP(new Temp.Temp());
	list.add(new Tree.MOVE(re, newSv));
*/
	Tree.SUBSCRIPTVAR newSv=new Tree.SUBSCRIPTVAR(sArray, sIndex, sWs);

//      Tree.TEMP re=new Tree.TEMP(new Temp.Temp());
//      list.add(new Tree.MOVE(re, newSv));

      return (Tree.Exp)newSv;
    }
    else {
      throw new FlatException("Simplifying an unknown Exp!");
    }
  }

  static public Tree.StmList flatten(Tree.StmList input) {
    Tree.Stm thestm;
    myStmList list=new myStmList();
    
    for (Tree.StmList curstm=input; curstm!=null; curstm=curstm.tail) {
      thestm=curstm.head;

      // check for poor statements
      if (thestm==null) {
        throw new FlatException("curstm.head is null!");
      }

      // LABEL, just add
      if (thestm instanceof Tree.LABEL) {
          list.add(thestm);
      }

      // SEQ, no SEQ in canonical tree!
      else if (thestm instanceof Tree.SEQ) {
          throw new FlatException("SEQ in canonical tree!");
      }

      // JUMP
      else if (thestm instanceof Tree.JUMP) {
          Tree.JUMP jump=(Tree.JUMP)thestm;

          Tree.Exp exp=simplifyExp(jump.exp, list); 

	  list.add(new Tree.JUMP(exp, jump.targets)); 
      }

      // CJUMP
      else if (thestm instanceof Tree.CJUMP) {
          Tree.CJUMP cjump=(Tree.CJUMP)thestm;

	  Tree.Exp left=simplifyExp(cjump.left, list);
	  Tree.Exp right=simplifyExp(cjump.right, list);

	  list.add(new Tree.CJUMP(cjump.relop, left, right, cjump.iftrue, cjump.iffalse)); 
      }

      // EXP
      else if (thestm instanceof Tree.EXP) {
          Tree.EXP instr=(Tree.EXP)thestm;

	    if (instr.exp instanceof Tree.CALL) {
	        Tree.CALL call=(Tree.CALL)(instr.exp);

	        if (call.func instanceof Tree.NAME) {
	            // build argument vector
		      Vector args=new Vector();

		      for (Tree.ExpList arg=call.args; arg!=null; arg=arg.tail) {
		          if (arg.head==null) {
		              throw new FlatException("arg.head is null!");
		          }
                      args.addElement(simplifyExp(arg.head, list));
		      }

		      // construct new call
		      if (call.args==null) {
		          list.add(new Tree.EXP(new Tree.CALL(call.func, null)));
		      } else {
		          Tree.ExpList argsList=new Tree.ExpList((Tree.Exp)args.lastElement(), null);
  		          for (int i=args.size()-2; i>=0; i--) {
		              argsList=new Tree.ExpList((Tree.Exp)(args.elementAt(i)), argsList);
		          }

		          list.add(new Tree.EXP(new Tree.CALL(call.func, argsList)));
		      }
	        }
	        else { // a NAME
	            throw new FlatException("Call with non-name function!");
	        }
	    } else {  // a CALL
	        throw new FlatException("No CALL in EXP statement!");
	    }
      }
     
      // MOVE
      else if (thestm instanceof Tree.MOVE) {
          Tree.MOVE move=(Tree.MOVE)thestm;

	  Tree.Exp src, dst;

          // note the evaluation order
	  src=simplifyExp(move.src, list);
	  dst=simplifyExp(move.dst, list);
          list.add(new Tree.MOVE(dst, src));
      }
    }

    /* convert myStmList to Tree.StmList */

    return list.toStmList();
  }
}
