/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package FlowGraph;
import Graph.Node;
import Graph.NodeList;
import Assem.Instr;

/**
 * AssemFlowGraph is an implementation of FlowGraph that constructs
 * a control flow graph based on a sequence of Assem instructions.
 * @see Assem.Instr
 * @see FlowGraph
 */

public class AssemFlowGraph extends FlowGraph {
     
 /**
  * Get the instruction associated with this node of the graph.
  */
     public Instr instr(Node n) {
	return ((FlowNode)n).instr;
     }

 /**
  * True if the instruction is a MOVE.
  * @see Assem.MOVE
  */
     public boolean isMove(Node n) {
	return ((FlowNode)n).instr instanceof Assem.MOVE;
     }

 /**
  * The set of temporaries defined by this instruction.
  */
     public Temp.TempList def(Node n) {
	return ((FlowNode)n).instr.def(); 
     }

 /**
  * The set of temporaries used by this instruction.
  */
     public Temp.TempList use(Node n) {
	return ((FlowNode)n).instr.use();
     }

     private java.util.Dictionary labmap = new java.util.Hashtable();

     Graph.NodeList orderedNodes;
 /**
  * A special version of the nodes() method that always returns the
  * nodes ordered the same way as the list of instructions passed
  * to the AssemFlowGraph constructor function.
  */
   public NodeList nodes() {return orderedNodes;}

     NodeList nextInstr(Assem.InstrList instrs) {
	if (instrs==null) return null;
	FlowNode n;
	NodeList tail = nextInstr(instrs.tail);
	FlowNode next = tail==null?  null : (FlowNode)tail.head;
	Assem.Instr instr = instrs.head;
	if (instr instanceof Assem.LABEL) {
	   n = (FlowNode)labmap.get(((Assem.LABEL)instr).label);
	   if (next!=null) addEdge(n, next);
	}
	else {
	   n = new FlowNode(this, instr);
	   if (instr.jumps()==null) {
	      if (next!=null) addEdge(n,next);
	   }
	   else
	     for(Temp.LabelList labs = instr.jumps().labels; 
					labs!=null; labs=labs.tail)
		addEdge(n,(FlowNode)labmap.get(labs.head));
	}
	return new NodeList(n,tail);
     }

 /**
  * Construct a FlowGraph from a list of Assem instructions.
  */
     public AssemFlowGraph(Assem.InstrList instrs) {
	int instrCount=0;
	for(Assem.InstrList p=instrs; p!=null; p=p.tail)
	    if (p.head instanceof Assem.LABEL)
	        labmap.put(((Assem.LABEL)p.head).label,
				      new FlowNode(this, p.head));
        orderedNodes = nextInstr(instrs);
     }

}

class FlowNode extends Node {
	Instr instr;
	FlowNode(Graph.Graph g, Instr i) {
	   super(g);
	   instr=i;
	}
}
