package Lower;

public class Lower {

  public static Tree.StmList lower(Tree.StmList sl) {
    if (sl == null) return null;
    return new Tree.StmList(lower(sl.head), lower(sl.tail));
  }

  static Tree.Stm lower(Tree.MOVE s) {
    return new Tree.MOVE(lower(s.dst), lower(s.src));
  }

  static Tree.Stm lower(Tree.EXP s) {
    return new Tree.EXP(lower(s.exp));
  }

  static Tree.Stm lower(Tree.JUMP s) {
    return new Tree.JUMP(lower(s.exp), lower(s.targets));
  }

  static Tree.Stm lower(Tree.CJUMP s) {
    return new Tree.CJUMP(s.relop,
			  lower(s.left),
			  lower(s.right),
			  lower(s.iftrue),
			  lower(s.iffalse));
  }

  static Tree.Stm lower(Tree.SEQ s) {
    return new Tree.SEQ(lower(s.left), lower(s.right));
  }

  static Tree.Stm lower(Tree.LABEL s) {
    return new Tree.LABEL(lower(s.label));
  }

  static Tree.Stm lower(Tree.Stm s) {
    if (s instanceof Tree.MOVE) return lower((Tree.MOVE)s);
    else if (s instanceof Tree.EXP) return lower((Tree.EXP)s);
    else if (s instanceof Tree.JUMP) return lower((Tree.JUMP)s);
    else if (s instanceof Tree.CJUMP) return lower((Tree.CJUMP)s);
    else if (s instanceof Tree.SEQ) return lower((Tree.SEQ)s);
    else if (s instanceof Tree.LABEL) return lower((Tree.LABEL)s);
    else throw new Error("Lower.lower.Stm");
  }

  static Tree.Exp lower(Tree.CONST e) {
    return new Tree.CONST(e.value);
  }

  static Tree.Exp lower(Tree.NAME e) {
    return new Tree.NAME(lower(e.label));
  }

  static Tree.Exp lower(Tree.TEMP e) {
    return new Tree.TEMP(lower(e.temp));
  }

  static Tree.Exp lower(Tree.BINOP e) {
    return new Tree.BINOP(e.binop, lower(e.left), lower(e.right));
  }

  static Tree.Exp lower(Tree.MEM e) {
    return new Tree.MEM(lower(e.exp));
  }

  static Tree.Exp lower(Tree.CALL e) {
    return new Tree.CALL(lower(e.func), lower(e.args));
  }

  static Tree.Exp lower(Tree.ESEQ e) {
    return new Tree.ESEQ(lower(e.stm), lower(e.exp));
  }

  static Tree.Exp lower(Tree.SIMPLEVAR e) {
    return lower(e.lowExp);
  }

  static Tree.Exp lower(Tree.FIELDVAR e) {
    return mem(plus(lower(e.record),
		    lower(e.offset)));
  }

  static Tree.Exp mem(Tree.Exp x) {return new Tree.MEM(x);}
  static Tree.Exp plus(Tree.Exp a, Tree.Exp b) {
    return new Tree.BINOP(Tree.BINOP.PLUS, a, b);
  }
  static Tree.Exp mul(Tree.Exp a, Tree.Exp b) {
    if (isconst(b,4)) 
          return new Tree.BINOP(Tree.BINOP.LSHIFT,a,new Tree.CONST(2));
    else  return new Tree.BINOP(Tree.BINOP.MUL,a,b);
  }
  static boolean isconst(Tree.Exp a, int c) {
      return a instanceof Tree.CONST && ((Tree.CONST)a).value==c;
  }

  static Tree.Exp lower(Tree.SUBSCRIPTVAR e) {
    return mem(plus(lower(e.array), 
		    mul(lower(e.index),
			lower(e.ws))));
  }

  static Tree.Exp lower(Tree.Exp e) {
    if (e instanceof Tree.CONST) return lower((Tree.CONST)e);
    else if (e instanceof Tree.NAME) return lower((Tree.NAME)e);
    else if (e instanceof Tree.TEMP) return lower((Tree.TEMP)e);
    else if (e instanceof Tree.BINOP) return lower((Tree.BINOP)e);
    else if (e instanceof Tree.MEM) return lower((Tree.MEM)e);
    else if (e instanceof Tree.CALL) return lower((Tree.CALL)e);
    else if (e instanceof Tree.ESEQ) return lower((Tree.ESEQ)e);
    else if (e instanceof Tree.SIMPLEVAR) return lower((Tree.SIMPLEVAR)e);
    else if (e instanceof Tree.FIELDVAR) return lower((Tree.FIELDVAR)e);
    else if (e instanceof Tree.SUBSCRIPTVAR) return 
					       lower((Tree.SUBSCRIPTVAR)e);
    else throw new Error("Lower.lower.Exp");
  }

  // helpers for things that aren't expressions or statements.
  // label should be as low as it gets
  static Temp.Label lower(Temp.Label l) {
    return l;
  }

  // labelList 
  static Temp.LabelList lower(Temp.LabelList ll) {
    if (ll == null) return null;
    return new Temp.LabelList(lower(ll.head), lower(ll.tail));
  }

  // temp should not get any lower
  static Temp.Temp lower(Temp.Temp t) {
    return t;
  }

  static Tree.ExpList lower(Tree.ExpList el) {
    if (el == null) return null;
    return new Tree.ExpList(lower(el.head), lower(el.tail));
  }
}
