/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Main;

import java.util.*;

class Main {

 static Frame.Frame frame = new Mips.MipsFrame();

 static Tree.StmList mergeStmList(Tree.StmList a, Tree.StmList b, Tree.StmList c) {
    Vector stms=new Vector();
    Tree.StmList r=null;

    for (Tree.StmList list=a; list!=null; list=list.tail) {
      stms.addElement(list.head);
    }
    for (Tree.StmList list=b; list!=null; list=list.tail) {
      stms.addElement(list.head);
    }
    for (Tree.StmList list=c; list!=null; list=list.tail) {
      stms.addElement(list.head);
    }

    for (int i=stms.size()-1; i>=0; i--) {
      r=new Tree.StmList((Tree.Stm)(stms.elementAt(i)), r);
    }

    return r;
 }

 static void prStmList(Tree.Print print, Tree.StmList stms) {
     for(Tree.StmList l = stms; l!=null; l=l.tail)
	print.prStm(l.head);
 }

 static Assem.InstrList codegen(Frame.Frame f, Tree.StmList stms) {
  	Assem.InstrList first=null, last=null;
	    for(Tree.StmList s=stms; s!=null; s=s.tail) {
	       Assem.InstrList i = f.codegen(s.head);
	       if (last==null) {first=last=i;}
	       else {while (last.tail!=null) last=last.tail;
		     last=last.tail=i;
		    }
	    }
	    return first;
	}
		

 static void emitParse(Absyn.Exp exp) {
     java.io.PrintStream debug;
     try {debug = new java.io.PrintStream(new java.io.FileOutputStream("absyn.out"));
	} catch (java.io.IOException e) {throw new Error(e.toString());}

     Absyn.Print print = new Absyn.Print(debug);
     print.prExp(exp, 0);
 }
 static void emitProc(java.io.PrintStream out, Translate.ProcFrag f) {
     java.io.PrintStream debug;
     try {debug = new java.io.PrintStream(new java.io.FileOutputStream(f.frame.name.toString() + ".out"));
	} catch (java.io.IOException e) {throw new Error(e.toString());}

     Temp.TempMap tempmap= new Temp.CombineMap(f.frame,new Temp.DefaultMap());
     Tree.Print print = new Tree.Print(debug, tempmap);
     debug.println("# Before canonicalization: ");
     print.prStm(f.body);

     debug.println("# After canonicalization: ");
     Tree.StmList cstms = Canon.Canon.linearize(f.body);
     Tree.StmList procstms = Canon.Canon.linearize(f.prologue);
     Tree.StmList epicstms = Canon.Canon.linearize(f.epilogue);
     prStmList(print,cstms);

     debug.println("# After flatten: ");
     Tree.StmList stms = Flat.Flat.flatten(cstms);
     Tree.StmList prostms = Flat.Flat.flatten(procstms);
     Tree.StmList epistms = Flat.Flat.flatten(epicstms);
     prStmList(print,stms);

     debug.println("# Basic Blocks: ");
     Canon.BasicBlocks b = new Canon.BasicBlocks(stms);
//     Canon.BasicBlocks prob = new Canon.BasicBlocks(prostms);
//     Canon.BasicBlocks epib = new Canon.BasicBlocks(epistms);
     for(Canon.StmListList l = b.blocks; l!=null; l=l.tail) {
       debug.println("#");
       prStmList(print,l.head);
     }       
     print.prStm(new Tree.LABEL(b.done));

     // do opts here...

     debug.println("# Trace Scheduled: ");
     Tree.StmList traced = (new Canon.TraceSchedule(b)).stms;
//    Tree.StmList protraced = (new Canon.TraceSchedule(prob)).stms;
//     Tree.StmList epitraced = (new Canon.TraceSchedule(epib)).stms;
     prStmList(print,traced);

     // should do lowering here
     debug.println("# After lowering: ");
     Tree.StmList bodyLowStms = Lower.Lower.lower(traced);
     Tree.StmList proLowStms = Lower.Lower.lower(prostms);
     Tree.StmList epiLowStms = Lower.Lower.lower(epistms);
/*     Tree.StmList proLowStms = Lower.Lower.lower(protraced);
     Tree.StmList epiLowStms = Lower.Lower.lower(epitraced);
*/    
     Tree.StmList lowStms=mergeStmList(proLowStms, bodyLowStms, epiLowStms);

     prStmList(print, lowStms);

     Assem.InstrList instrs= codegen(f.frame,lowStms);
     //     Assem.InstrList instrs= codegen(f.frame,traced);
     instrs = f.frame.procEntryExit2(instrs);
     debug.println("# Instructions: ");
     for(Assem.InstrList p=instrs; p!=null; p=p.tail)
	debug.print(p.head.format(tempmap));
     debug.println("# Register Allocated: ");
     RegAlloc.RegAlloc regA = new RegAlloc.RegAlloc(f.frame,instrs);
     Frame.Proc proc = f.frame.procEntryExit3(regA.instrs);
     out.print(proc.prolog);
     for(Assem.InstrList p=proc.body; p!=null; p=p.tail)
	out.print(p.head.format(regA));
     out.print(proc.epilog);
     debug.close();
 }

 public static void main(String args[]) throws java.io.IOException {
   Parse.Parse parse = new Parse.Parse(args[0]);
   java.io.PrintStream out = 
        new java.io.PrintStream(new java.io.FileOutputStream(args[0].substring(0, args[0].length() - 4) + ".s"));
   Translate.Translate translate = new Translate.Translate(frame);
   Semant.Semant semant = new Semant.Semant(translate,parse.errorMsg);
   Translate.Frag frags = semant.transProg(parse.absyn);
   emitParse(parse.absyn);
   for(Translate.Frag f = frags; f!=null; f=f.next)
       if (f instanceof Translate.ProcFrag)
          emitProc(out, (Translate.ProcFrag)f);
       else if (f instanceof Translate.DataFrag)
          out.print(((Translate.DataFrag)f).data);
   out.close();
 }

}

class NullOutputStream extends java.io.OutputStream {
    public void write(int b) {}
}


