/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Mips;
import Assem.OPER;
import Assem.MOVE;
import Assem.LABEL;
import Temp.Temp;
import Temp.TempList;
import Temp.Label;
import Temp.LabelList;

public class Codegen {

  MipsFrame frame;
  String framename;
  TempList calldefs;

  public Codegen(MipsFrame f) {
	frame=f;  
	framename=frame.name.toString();
 	calldefs = cons(frame.RV(), cons(frame.RAx, frame.callerSaves));
  }

  static TempList cons(Temp h, TempList t) {
	return new TempList(h,t);
  }

  Assem.InstrList ilist=null, last=null;

  void emit(Assem.Instr inst) {
     if (last!=null) 	
	last = last.tail = new Assem.InstrList(inst,null);
     else last = ilist = new Assem.InstrList(inst,null);
  }

  void doStm(Tree.SEQ s) {
	doStm(s.left); doStm(s.right);
  }
  void doStm(Tree.LABEL s) {
	emit(new LABEL(s.label.toString() + ":\n", s.label));
  }
  void doStm(Tree.JUMP s) {
     if (s.exp instanceof Tree.NAME) 
        emit(new OPER("\tb\t`j0\n", null, null, s.targets));
     else 
	emit(new OPER("\tj\t`s0\n", null, new TempList(doExp(s.exp),null),
	     s.targets));
  }
  void doStm(Tree.CJUMP s) {
    if (s.relop==Tree.CJUMP.EQ) 
	equal("eq",s.left,s.right,s.iftrue,s.iffalse);
    else if (s.relop==Tree.CJUMP.NE) 
	equal("ne",s.left,s.right,s.iftrue,s.iffalse);
    else {
	Test t = test(s.relop,s.left,s.right);
        emit(new OPER("\tb" + t.cond + "\t`s0,$0,`j0\n",
		      null, new TempList(t.temp,null),
		      new LabelList(s.iftrue, 
			new LabelList(s.iffalse, null))));
    }
  }
  void doStm(Tree.MOVE s) {
     if (s.dst instanceof Tree.TEMP)
	   move(((Tree.TEMP)s.dst).temp, doExp(s.src));
     else if (s.dst instanceof Tree.MEM)
	   mem(((Tree.MEM)s.dst).exp, true, doExp(s.src));
     else throw new Error("bad MOVE in Mips.Codegen");
  }

  void doStm(Tree.EXP s) {
      doExp(s.exp);
  }

  void doStm(Tree.Stm s) {
          if (s instanceof Tree.SEQ) doStm((Tree.SEQ)s);
     else if (s instanceof Tree.LABEL) doStm((Tree.LABEL)s);
     else if (s instanceof Tree.JUMP) doStm((Tree.JUMP)s);
     else if (s instanceof Tree.CJUMP) doStm((Tree.CJUMP)s);
     else if (s instanceof Tree.MOVE) doStm((Tree.MOVE)s);
     else if (s instanceof Tree.EXP) doStm((Tree.EXP)s);
     else throw new Error("bad Stm in Mips.Codegen");
  }
        
  boolean isFP(Tree.Exp e) {
       return e instanceof Tree.TEMP && ((Tree.TEMP)e).temp==frame.FP();
  }

  Temp doExp(Tree.BINOP e) {
      switch(e.binop) {
        case Tree.BINOP.MINUS:
		if (e.right instanceof Tree.CONST)
		    return binopc("add",e.left,
			       new Tree.CONST(-((Tree.CONST)e.right).value));
		else return binop("sub",e.left,e.right);
        case Tree.BINOP.PLUS:
	        if (e.left instanceof Tree.CONST && isFP(e.right)) {
		     Temp x = new Temp();
		     emit(new OPER("\taddiu\t`d0,$sp," 
				   + String.valueOf(((Tree.CONST)e.left).value)
				   + "+" + framename + "_framesize\n",
				   new TempList(x,null),null));
		     return x;
		}
		else return binopc("add",e.left,e.right);
        case Tree.BINOP.AND:
		return binopc("and",e.left,e.right);
        case Tree.BINOP.OR:
		return binopc("or",e.left,e.right);
        case Tree.BINOP.XOR:
		return binopc("xor",e.left,e.right);
        case Tree.BINOP.LSHIFT:
		return shift("ll",e.left,e.right);
        case Tree.BINOP.RSHIFT:
		return shift("rl",e.left,e.right);
        case Tree.BINOP.ARSHIFT:
		return shift("ra",e.left,e.right);
        case Tree.BINOP.MUL:
		return muldiv("mult",e.left,e.right);
        case Tree.BINOP.DIV:
		return muldiv("div",e.left,e.right);
	default: throw new Error("bad BINOP in Mips.Codegen");
       }
   }

  Temp doExp(Tree.TEMP e) {
	if (isFP(e))
	   return doExp(new Tree.BINOP(Tree.BINOP.PLUS,new Tree.CONST(0),
				       new Tree.TEMP(frame.FP())));
	else return e.temp;
  }

  Temp doExp(Tree.MEM e) {
	Temp x = new Temp();
	mem(e.exp,false,x);
	return x;
  }

  Temp doExp(Tree.NAME e) {
	Temp x = new Temp();
	emit(new OPER("\tla\t`d0," + e.label.toString() + "\n",
		      new TempList(x,null), null));
	return x;
  }

  Temp doExp (Tree.CONST i) {
    return constant(i.value);
  }

  Temp doExp(Tree.CALL e) {
	if (e.func instanceof Tree.NAME) {
	   Label lab = ((Tree.NAME)e.func).label;
	   emit(new OPER("\tjal\t" + lab.toString() + "\n",
			 calldefs, argbuild(e.args)));
	}
	else 
	   emit(new OPER("\tjal\t`s0\n",
			 calldefs, new TempList(doExp(e.func),
					argbuild(e.args))));
	return frame.RV();
  }

  Temp doExp(Tree.Exp e) {
	     if (e instanceof Tree.BINOP) return doExp((Tree.BINOP)e);
	else if (e instanceof Tree.MEM) return doExp((Tree.MEM)e);
	else if (e instanceof Tree.NAME) return doExp((Tree.NAME)e);
	else if (e instanceof Tree.TEMP) return doExp((Tree.TEMP)e);
	else if (e instanceof Tree.CONST) return doExp((Tree.CONST)e);
	else if (e instanceof Tree.CALL) return doExp((Tree.CALL)e);
        else throw new Error("bad Exp in Mips.Codegen");
   }

   TempList argbuild(Tree.ExpList args) {
	return argbuild(0, frame.callerSaves, args);
   }

   TempList argbuild(int i, TempList regs, Tree.ExpList args) {
      if (args==null) return null;
      else if (i < frame.numArgRegs) {
	doStm(new Tree.MOVE(new Tree.TEMP(regs.head), args.head));
	return new TempList(regs.head, argbuild(i+1,regs.tail,args.tail));
      }
      else {
	doStm(new Tree.MOVE(new Tree.MEM(new Tree.BINOP(Tree.BINOP.PLUS,
				new Tree.CONST(i*frame.wordSize()),
				new Tree.TEMP(frame.FP()))),
			    args.head));
	return argbuild(i+1,null,args.tail);
      }
    }

   String positiveop(boolean b) {
	return b ? "eq" : "ne";
   }

   Test simpletest(boolean positive, String unsigned, Tree.Exp a, Tree.Exp b,
			boolean reverse) {
     String ab = reverse ? "`s1,`s0" : "`s0,`s1";
     Temp x = new Temp();
     emit(new OPER("\tslt" + unsigned + "\t`d0," + ab + "\n",
	           new TempList(x,null), 
		   new TempList(doExp(a), new TempList(doExp(b), null))));
     return new Test(positiveop(positive),x);
   }

   Test testx(boolean positive, String unsigned, Tree.Exp a, Tree.Exp b, 
		boolean reverse) {
	if (b instanceof Tree.CONST) {
	   int i = ((Tree.CONST)b).value;
	   if (reverse) {
	       if (i < Integer.MAX_VALUE)
		  return testx(!positive, unsigned, a, new Tree.CONST(i+1),
				false);
	       else return simpletest(positive,unsigned,a,b,reverse);
	   }
	   else {
  	       Temp x = new Temp();
	       emit(new OPER("\tslti" + unsigned + "\t`d0,`s0," + 
			     String.valueOf(i) + "\n",
			     new TempList(x,null), new TempList(doExp(a),null)));
	       return new Test(positiveop(positive), x);
	   }
	}
	else if (a instanceof Tree.CONST)
	    return testx(positive,unsigned,b,a,!reverse);
	else return simpletest(positive,unsigned,a,b,reverse);
   }

   Test test(int relop, Tree.Exp a, Tree.Exp b) {
       switch(relop) {
	case Tree.CJUMP.LT:  return testx(false,"", a,b,false);
	case Tree.CJUMP.GT:  return testx(false,"", a,b,true);
	case Tree.CJUMP.LE:  return testx(true, "", a,b,true);
	case Tree.CJUMP.GE:  return testx(true, "", a,b,false);
	case Tree.CJUMP.ULT: return testx(false,"u",a,b,false);
	case Tree.CJUMP.UGT: return testx(false,"u",a,b,true);
	case Tree.CJUMP.ULE: return testx(true, "u",a,b,true);
	case Tree.CJUMP.UGE: return testx(true, "u",a,b,false);
	default: throw new Error("bad RELOP in Mips.CODEGEN");
       }
    }

   void equal(String cond, Tree.Exp a, Tree.Exp b, Label t, Label f){
	emit(new OPER("\tb" + cond + "\t`s0,`s1,`j0\n",
		      null, new TempList(doExp(a), 
				new TempList(doExp(b),null)),
		      new LabelList(t, new LabelList(f, null))));
   }

   Temp constant(int i) {
	if (i==0) return frame.ZEROx;
	else if (-32768 < i && i < 32768) {
	   Temp x = new Temp();
	   emit (new OPER("\taddi\t`d0,$0," + String.valueOf(i) + "\n",
			  new TempList(x,null), null));
	   return x;
	}
	else {
	   Temp x = new Temp(), y = new Temp();
	   int hi =  (i>>16) & 0xffff;
	   int lo = i & 0xffff;
	   emit(new OPER("\tlui\t`d0," + String.valueOf(hi) + "\n",
			 new TempList(y,null), null));
	   emit(new OPER("\tori\t`d0,`s0," + String.valueOf(lo) + "\n",
			  new TempList(x,null), new TempList(y,null)));
	   return x;
	}
     }

  void mem(Tree.BINOP e, boolean isStore, Temp r) {
	if (e.binop==Tree.BINOP.PLUS && e.right instanceof Tree.CONST)
	  memx(e.left,((Tree.CONST)e.right).value, isStore, r);
	else if (e.binop==Tree.BINOP.PLUS && e.left instanceof Tree.CONST)
	  memx(e.right,((Tree.CONST)e.left).value, isStore, r);
	else if (e.binop==Tree.BINOP.MINUS && e.right instanceof Tree.CONST)
	  memx(e.left,(-((Tree.CONST)e.right).value), isStore, r);
	else 
	  memx(e,0, isStore, r);
  }

  void mem(Tree.Exp e, boolean isStore, Temp r) {
	if (e instanceof Tree.BINOP)
	  mem((Tree.BINOP)e, isStore, r);
	else memx(e,0,isStore,r);
  }

  void memx(Tree.Exp e, int offset, boolean isStore, Temp r) {
	if (-32768 < offset && offset < 32768) {
	   String sizecode = "w";
	   String i = String.valueOf(offset);
	   Temp ex;
	   if (isFP(e)) {
	         i = i + "+" + framename + "_framesize";
		 ex = frame.SPx;
	   }
	   else ex = doExp(e);
	   if (isStore)
		emit(new OPER("\ts"+sizecode+"\t`s1,"+i+"(`s0)\n",
			      null, new TempList(ex, new TempList(r, null))));
	   else emit(new OPER("\tl"+sizecode+"\t`d0,"+i+"(`s0)\n",
			      new TempList(r,null), new TempList(ex,null)));
	}
	else {
	   int lo = offset & 0xffff;
	   int hi = (offset >> 16) & 0xffff;
	   if (lo >= 0x8000) {
		lo = -(0x10000-lo);
		hi++;
	   }
	   memx(new Tree.BINOP(Tree.BINOP.PLUS,e,new Tree.CONST(hi)),
		lo,isStore,r);
	}
    }

  Temp muldiv(String instr, Tree.Exp a, Tree.Exp b) {
	Temp x = new Temp();
	emit(new OPER("\t"+instr+"\t`s0,`s1\n\tmflo\t`d0\n",
		  new TempList(x,null),
		  new TempList(doExp(a), new TempList(doExp(b),null))));
	return x;
  }

  Temp shift(String dir, Tree.Exp a, Tree.Exp b) {
	Temp x = new Temp();
	if (b instanceof Tree.CONST) {
	   String i = String.valueOf(((Tree.CONST)b).value);
	   emit(new OPER("\ts"+dir+"\t`d0,`s0,"+i+"\n",
			 new TempList(x,null), new TempList(doExp(a),null)));
	}
	else emit(new OPER("\ts" + dir + "v\t`d0,`s0,`s1\n", 
			 new TempList(x,null), 
			 new TempList(doExp(a), new TempList(doExp(b),null))));		  
	return x;
   }

  Temp binopc(String instr, Tree.Exp a, Tree.Exp b) {
	if (b instanceof Tree.CONST) {
	   Temp x = new Temp();
	   String i = String.valueOf(((Tree.CONST)b).value);
	   emit(new OPER("\t" + instr + "i\t`d0,`s0," + i + "\n", 
			 new TempList(x,null), 
			 new TempList(doExp(a), null)));
	   return x;
	}
	else if (a instanceof Tree.CONST)
	   return binopc(instr,b,a);
	else return binop(instr,a,b);
   }

   Temp binop(String instr, Tree.Exp a, Tree.Exp b) {
	Temp x = new Temp();
	emit(new OPER("\t"+instr+"\t`d0,`s0,`s1\n",
			 new TempList(x,null), 
			 new TempList(doExp(a), new TempList(doExp(b),null))));
        return x;
   }

	
   void move(Temp dst, Temp src) {
	emit(new MOVE("\tor\t`d0,$0,`s0\n", dst,src));
   }

   Assem.InstrList codegen(Tree.Stm s) {
	Assem.InstrList l;
	doStm(s);
	l=ilist;
	ilist=last=null;
	return l;
   }

}


class Test {
	String cond;
	Temp temp;
	Test(String c, Temp t) {cond=c; temp=t;}
}
