/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Parse;
	
public class HandLexer implements Lexer {

  java.io.InputStream inp;
  ErrorMsg.ErrorMsg errorMsg;

  public HandLexer(java.io.InputStream i, ErrorMsg.ErrorMsg err) { 
	errorMsg=err;
	inp=i; advance(); 
  }
  
  private int next_char;
  private int leftpos, pos;

  private java_cup.runtime.Symbol tok(int kind) {
    return new java_cup.runtime.Symbol(kind, leftpos, pos);
  }

  private void advance()  { try {next_char = inp.read();
				 pos++;
			    } catch (java.io.IOException e) {
				throw new Error("I/O Exception in Lexer");
			    }
			  }
			  

  private void newline() {
	errorMsg.newline(pos);
  }

  final static java.util.Hashtable keywordTable = new java.util.Hashtable();

  static {
	keywordTable.put("for",new Integer(sym.FOR));
	keywordTable.put("while",new Integer(sym.WHILE));
	keywordTable.put("break",new Integer(sym.WHILE));
	keywordTable.put("let",new Integer(sym.LET));
	keywordTable.put("in",new Integer(sym.IN));
	keywordTable.put("nil",new Integer(sym.NIL));
	keywordTable.put("to",new Integer(sym.TO));
	keywordTable.put("end",new Integer(sym.END));
	keywordTable.put("function",new Integer(sym.FUNCTION));
	keywordTable.put("var",new Integer(sym.VAR));
	keywordTable.put("type",new Integer(sym.TYPE));
	keywordTable.put("array",new Integer(sym.ARRAY));
	keywordTable.put("if",new Integer(sym.IF));
	keywordTable.put("then",new Integer(sym.THEN));
	keywordTable.put("else",new Integer(sym.ELSE));
	keywordTable.put("do",new Integer(sym.DO));
	keywordTable.put("of",new Integer(sym.OF));
  }

  /* recognize and return the next complete token */
  public java_cup.runtime.Symbol nextToken()
    {
      for (;;) {
        leftpos=pos;
        switch (next_char)
          {
            case '0': case '1': case '2': case '3': case '4': 
            case '5': case '6': case '7': case '8': case '9': 
              /* parse a decimal integer */
              int i_val = 0;
              do {
                i_val = i_val * 10 + (next_char - '0');
                advance();
              } while (next_char >= '0' && next_char <= '9');
            return new java_cup.runtime.Symbol(sym.INT,leftpos, pos,
						new Integer(i_val));

	    case ',': advance(); return tok(sym.COMMA);
	    case '{': advance(); return tok(sym.LBRACE);
	    case '}': advance(); return tok(sym.RBRACE);
	    case '[': advance(); return tok(sym.LBRACK);
	    case ']': advance(); return tok(sym.RBRACK);
	    case ':': advance(); 
			if (next_char=='=') {
			  advance(); return tok(sym.ASSIGN);
			} 
			else return tok(sym.COLON);
	    case ';': advance(); return tok(sym.SEMICOLON);
	    case '(': advance(); 
			if (next_char=='*') {
			  advance(); getComment(); break;
			} 
			else return tok(sym.LPAREN);
	    case ')': advance(); return tok(sym.RPAREN);
	    case '.': advance(); return tok(sym.DOT);
	    case '+': advance(); return tok(sym.PLUS);
	    case '-': advance(); return tok(sym.MINUS);
	    case '*': advance(); return tok(sym.TIMES);
	    case '/': advance(); return tok(sym.DIVIDE);
	    case '&': advance(); return tok(sym.AND);
	    case '|': advance(); return tok(sym.OR);
	    case '=': advance(); return tok(sym.EQ);
	    case '>': advance(); 
			if (next_char=='=') {
			  advance(); return tok(sym.GE);
			} 
			else return tok(sym.GT);
	    case '<': advance(); 
			if (next_char=='=') {
			  advance(); return tok(sym.LE);
			} 
			else return tok(sym.LT);

	    case '"': advance();
			return getString();

	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
	    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
	    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
	    case 'y': case 'z': case 'A': case 'B': case 'C': case 'D':
	    case 'E': case 'F': case 'G': case 'H': case 'I': case 'J':
	    case 'K': case 'L': case 'M': case 'N': case 'O': case 'P':
	    case 'Q': case 'R': case 'S': case 'T': case 'U': case 'V':
	    case 'W': case 'X': case 'Y': case 'Z':        
	      {StringBuffer b = new StringBuffer();
	       do {
		 b.append((char)next_char);
		 advance();
	       } while (next_char >= 'a' && next_char <= 'z'  ||
		        next_char >= 'A' && next_char <= 'Z'  ||
		        next_char >= '0' && next_char <= '9' );
	       {String s = b.toString();
	        Integer t = (Integer)keywordTable.get(s);
	        if (t != null) return tok(t.intValue());
	        return new java_cup.runtime.Symbol(sym.ID, leftpos, pos, s);
	      }}


            case -1: return tok(sym.EOF);

	    case ' ': case '\t':
	      advance();
	      break;

	    case '\n': 
	      advance(); newline();
	      break;

            default:

	      errorMsg.error(0,"illegal token");
              advance();
       }  }
    }

    private java_cup.runtime.Symbol getString() {
	StringBuffer b = new StringBuffer();
	for(;;)
	  switch (next_char) {
	    case '"': {java_cup.runtime.Symbol t = new java_cup.runtime.Symbol(sym.STRING,
					leftpos, pos, b.toString());
	               advance();	
		       return t;
		      }
	    case '\\':
		advance();
		switch (next_char) {
		  case 'n': advance(); b.append('\n'); break;
		  case 't': advance(); b.append('\t'); break;
		  default:
		      errorMsg.error(0,"unrecognized string escape");
		}
		break;
 	    case '\n': 
		      advance(); newline();
		      errorMsg.error(0,"unescaped newline in string");
	    case -1:
		      errorMsg.error(0,"unclosed string");
		      return new java_cup.runtime.Symbol(sym.STRING,leftpos, pos,b.toString());
	    default:
	      b.append((char)next_char);
	      advance();
	  }
    }

   private void getComment() {
	for(;;)
	  switch (next_char) {
	     case '*': advance(); 
		       if (next_char==')')  {
			   advance();
			   return;
		       }
	     case '(': advance();
			if (next_char=='*')
			   getComment();
		       break;
	     case -1: 
		      errorMsg.error(0,"unclosed comment");
		      return;
	     case '\n':
			advance();
			newline();
			break;
	     default:
		advance();
	  }
   }
			  

};

