/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Parse;

%% 

%implements Lexer
%function nextToken
%type java_cup.runtime.Symbol
%char

%{
private int stringstart;

private StringBuffer charbuf = null;

private void newline() {
  errorMsg.newline(yychar);
}

private int comLevel = 0;

private void err(int pos, String s) {
  errorMsg.error(pos,s);
}

private void err(String s) {
  err(yychar,s);
}

private java_cup.runtime.Symbol tok(int kind) {
    return new java_cup.runtime.Symbol(kind, yychar, yychar+yylength());
}


private ErrorMsg.ErrorMsg errorMsg;

Yylex(java.io.InputStream s, ErrorMsg.ErrorMsg e) {
  this(s);
  errorMsg=e;
}

%}

%eofval{
	{if (comLevel>0) err (stringstart, "unclosed comment");
	 return tok(sym.EOF);
        }
%eofval}       


%state A S F
idchars=[A-Za-z_0-9]
id=[A-Za-z]{idchars}*
ws=("\012"|[\t\ ])*
%%
<YYINITIAL>{ws}	{}
<YYINITIAL>\n	{newline();}
<YYINITIAL>","	{return tok(sym.COMMA);}
<YYINITIAL>"{"	{return tok(sym.LBRACE);}
<YYINITIAL>"}"	{return tok(sym.RBRACE);}
<YYINITIAL>"["	{return tok(sym.LBRACK);}
<YYINITIAL>"]"	{return tok(sym.RBRACK);}
<YYINITIAL>":"	{return tok(sym.COLON);}
<YYINITIAL>";"	{return tok(sym.SEMICOLON);}
<YYINITIAL>"("	{return tok(sym.LPAREN);}
<YYINITIAL>")"	{return tok(sym.RPAREN);}
<YYINITIAL>"."	{return tok(sym.DOT);}
<YYINITIAL>"+"	{return tok(sym.PLUS);}
<YYINITIAL>"-"	{return tok(sym.MINUS);}
<YYINITIAL>"*"	{return tok(sym.TIMES);}
<YYINITIAL>"/"	{return tok(sym.DIVIDE);}
<YYINITIAL>"&"	{return tok(sym.AND);}
<YYINITIAL>"|"	{return tok(sym.OR);}
<YYINITIAL>":="	{return tok(sym.ASSIGN);}
<YYINITIAL>"="	{return tok(sym.EQ);}
<YYINITIAL>"<>"	{return tok(sym.NEQ);}
<YYINITIAL>">"	{return tok(sym.GT);}
<YYINITIAL>"<"	{return tok(sym.LT);}
<YYINITIAL>">="	{return tok(sym.GE);}
<YYINITIAL>"<="	{return tok(sym.LE);}
<YYINITIAL>for	{return tok(sym.FOR);}
<YYINITIAL>while {return tok(sym.WHILE);}
<YYINITIAL>break {return tok(sym.WHILE);}
<YYINITIAL>let 	{return tok(sym.LET);}
<YYINITIAL>in  	{return tok(sym.IN);}
<YYINITIAL>nil 	{return tok(sym.NIL);}
<YYINITIAL>to  	{return tok(sym.TO);}
<YYINITIAL>end 	{return tok(sym.END);}
<YYINITIAL>function  {return tok(sym.FUNCTION);}
<YYINITIAL>var 	{return tok(sym.VAR);}
<YYINITIAL>type	{return tok(sym.TYPE);}
<YYINITIAL>array  {return tok(sym.ARRAY);}
<YYINITIAL>if 	{return tok(sym.IF);}
<YYINITIAL>then {return tok(sym.THEN);}
<YYINITIAL>else {return tok(sym.ELSE);}
<YYINITIAL>do  	{return tok(sym.DO);}
<YYINITIAL>of  	{return tok(sym.OF);}
<YYINITIAL>{id}	{return new java_cup.runtime.Symbol(sym.ID,
				yychar,yychar+yylength(),
				     yytext());}
<YYINITIAL>[0-9]+ {return new java_cup.runtime.Symbol(sym.INT,
			yychar,yychar+yylength(),
			  new Integer(Integer.parseInt(yytext())));}
<YYINITIAL>\"	{charbuf=new StringBuffer(); stringstart=yychar; yybegin(S);}
<YYINITIAL>"/*"	{yybegin(A); stringstart = yychar; comLevel = 1;}
<YYINITIAL>"*/"	{err("unmatched close comment");}
<YYINITIAL>\h	{err("non-Ascii character");}
<YYINITIAL>.	{err("illegal token");}
<A>"/*"		{comLevel++;}
<A>\n		{newline();}
<A>"*/" 	{comLevel--; if (comLevel==0) yybegin(YYINITIAL);}
<A>.		{}
<S>\"	        {yybegin(YYINITIAL);
		 return new java_cup.runtime.Symbol(sym.STRING,
			stringstart,yychar+yylength(),charbuf.toString());
		}
<S>\n		{err (stringstart,"unclosed string");
		 newline();
		 yybegin(YYINITIAL); 
		 return new java_cup.runtime.Symbol(sym.STRING,
				stringstart,yychar+yylength(),
				     charbuf.toString());
		}
<S>\\\n	       	{newline(); yybegin(F);}
<S>\\[\ \t]   	{yybegin(F);}
<F>\n		{newline();}
<F>{ws}		{}
<F>\\		{yybegin(S); stringstart=yychar;}
<F>.		{err (stringstart,"unclosed string"); yybegin(YYINITIAL); 
		 return new java_cup.runtime.Symbol(sym.STRING,
			stringstart,yychar+yylength(),
				     charbuf.toString());
		}
<S>\\t		{charbuf.append('\t');}
<S>\\n		{charbuf.append('\n');}
<S>\\\\		{charbuf.append('\\');}
<S>\\\"		{charbuf.append('\"');}
<S>\\"^"[@-_]	{charbuf.append((char)(yytext().charAt(2) - '@'));}
<S>\\[0-9][0-9][0-9]	{String text = yytext();
			 int x = text.charAt(1)*100 + text.charAt(2)*10 
		                     + text.charAt(3) - ('0' * 111);
		 if (x>255)
		      err("illegal ascii escape");
		 else charbuf.append((char)x);
		}
<S>\\		{err("illegal string escape"); }
<S>.		{charbuf.append(yytext());}


