/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package RegAlloc;
import Graph.Node;
import Graph.NodeList;
import Temp.Temp;
import Temp.TempList;
import FlowGraph.FlowGraph;

public class Liveness extends InterferenceGraph {
   
   java.util.Dictionary temp2node = new java.util.Hashtable();

   java.util.Dictionary livemap = new java.util.Hashtable();

   public Node tnode(Temp temp) {
	return (Node)temp2node.get(temp);
   }
   public Temp gtemp(Node node) {
	return ((TempNode)node).temp;
   }

   MoveList theMoves;
   public MoveList moves() {
	return theMoves;
   }
   void makeNodes(TempList temps) {
	  for(TempList l = temps; l!=null; l=l.tail)
	    if (tnode(l.head)==null)
		new TempNode(this, l.head);
   }

   void trace(FlowGraph flow, TempNode tn, Node flownode) {
	  /* Precondition: tn is live at entry to node
	     Postcondition: all consequences of that fact backwards
	            to (but not through) any def have been incorporated
		    into livemap and interference graph
	  */
      TempList livehere = (TempList)livemap.get(flownode);
      livemap.put(flownode, new TempList(gtemp(tn),livehere));
      if ( ! tn.inList(livehere))
	for(NodeList preds = flownode.pred(); preds!=null; preds=preds.tail) {
	   TempList defs = flow.def(preds.head);
	   for(TempList d = defs; d!=null; d=d.tail)
	       if (tn != tnode(d.head))
		   addEdge(tn, tnode(d.head));
	   if (!tn.inList(defs))
		trace(flow, tn, preds.head);
        }

    }
	
   void traceOneTemp(FlowGraph flow, TempNode tn) {
   /* Precondition: temp not member of liveMap(node) for any node.
      Postcondition: 1. for any n, temp in live(n) iff temp is
                          live on ENTRY to n.
		     2. for any (t,m,n), iff t in def(m)  &  temp in live(n)
                              & m in pred(n) & t <> temp  
                         then addEdge(t,temp) has been called
   */
   
     for(NodeList p = flow.nodes(); p!=null; p=p.tail)
	if (tn.inList(flow.use(p.head)))
	   trace(flow,tn,p.head);
   }

   public Liveness(FlowGraph flow) {
	for(NodeList p = flow.nodes(); p!=null; p=p.tail) {
            makeNodes(flow.use(p.head));
	    makeNodes(flow.def(p.head));
        }
	for (NodeList tp = nodes(); tp!=null; tp=tp.tail)
	   traceOneTemp(flow, (TempNode)tp.head);
   }

}

class TempNode extends Node {
  Temp temp;
  TempNode(Liveness g, Temp t) {
	super(g);
	temp=t;
	g.temp2node.put(t, this);

  }
  public String toString() {return temp.toString();}
  boolean inList(TempList temps) {
	if (temps==null) return false;
	else if (temps.head==temp) return true;
	else return inList(temps.tail);
  }
}
