/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package RegAlloc;
import Assem.Instr;
import Assem.InstrList;
import Temp.TempList;
import Temp.Temp;
import Frame.Frame;
import Frame.Access;

class Replacement {
	TempList temps;
	InstrList instrs;
	Replacement(TempList t, InstrList i) {temps=t; instrs=i;}
}

class Spill {

  InstrList instrs=null;

  Frame frame;

  java.util.Dictionary locmap = new java.util.Hashtable();

  Access locOf(Temp t) {return (Access)locmap.get(t);}

  InstrList splice(InstrList a, InstrList b) {
	if (a==null) return b;
        InstrList p;
	for(p=a; p.tail!=null; p=p.tail) ;
	p.tail=b;
	return a;
  }
	

  InstrList replace(boolean isUse, TempList temps) {
	InstrList instrs = null;
	for(TempList p = temps; p!=null; p=p.tail) {
	   Access a = locOf(p.head);
	   if (a!=null) {
	      Temp u = new Temp();
	      Tree.Exp memloc = a.exp(new Tree.TEMP(frame.FP()));
	      Tree.Exp ue = new Tree.TEMP(u);
	      Tree.Stm move = isUse? new Tree.MOVE(ue,memloc)
				   : new Tree.MOVE(memloc,ue);
	      InstrList newInstrs = frame.codegen(move);
	      instrs = splice(newInstrs, instrs);
	      p.head=u;
	   }
	}
	return instrs;
  }

  

  InstrList augment(InstrList il, Assem.OPER i) {
	InstrList src = replace(true,i.src);
	InstrList dst = replace(false,i.dst);
	return splice(src,splice(il,dst));
  }	
	
  InstrList augment(InstrList il, Assem.MOVE i) {
	TempList s = new TempList(i.src,null);
	InstrList src = replace(true,s);
	i.src=s.head;
	s.head=i.dst;
	InstrList dst = replace(false,s);
	i.dst=s.head;
	return splice(src,splice(il,dst));
  }	
	
  InstrList augment(InstrList il, Instr i) {
	if (i instanceof Assem.OPER) return augment(il, (Assem.OPER)i);
	if (i instanceof Assem.MOVE) return augment(il, (Assem.MOVE)i);
	if (i instanceof Assem.LABEL) return il;
	throw new Error("Spill.augment");
  }
	


  Spill(Frame f, InstrList il, TempList spills) {
	frame=f;
	for(TempList p=spills; p!=null; p=p.tail)
	  locmap.put(p.head,frame.allocLocal(true));
	InstrList preHead = new InstrList(null,il);
	InstrList prev=preHead, p = preHead.tail;
	while (p!=null) {
	   InstrList next = p.tail;
	   p.tail=null;
	   prev.tail = augment(p, p.head);
	   while (p.tail!=null) p=p.tail;
	   p.tail = next;
	   prev = p;
	   p = next;
	   
	}
	instrs = preHead.tail;
     }


}
