/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Translate;
import Temp.Label;	
import Temp.Temp;
import java.util.*;

class Ex extends Exp {
  Tree.Exp exp;
  Ex(Tree.Exp e) {exp=e;}
  Tree.Exp unEx() {return exp;}
  Tree.Stm unCx(Label t, Label f) {
      if (exp instanceof Tree.CONST)
        if (((Tree.CONST)exp).value!=0)
             return new Tree.JUMP(t);
        else return new Tree.JUMP(f); 
      else return new Tree.CJUMP(Tree.CJUMP.NE,exp,new Tree.CONST(0),t,f);
  }
  Tree.Stm unNx() {return new Tree.EXP(exp);}
}

class Nx extends Exp {
  Tree.Stm stm;
  Nx(Tree.Stm s) {stm=s;}
  Tree.Exp unEx() {return new Tree.ESEQ(stm, new Tree.CONST(0));}
  Tree.Stm unCx(Label t, Label f) {
	throw new Error("Nx.unCx");
  }
  Tree.Stm unNx() {return stm;}
}

abstract class Cx extends Exp {
  Tree.Exp unEx() {
    Temp r = new Temp();
    Label t = new Label();
    Label f = new Label();

    return new Tree.ESEQ(new Tree.SEQ(new Tree.MOVE(new Tree.TEMP(r), 
				           new Tree.CONST(1)),
		      new Tree.SEQ(unCx(t,f),
		       new Tree.SEQ(new Tree.LABEL(f),
		        new Tree.SEQ(new Tree.MOVE(new Tree.TEMP(r), new Tree.CONST(0)),
			 new Tree.LABEL(t))))),
		     new Tree.TEMP(r));
 }

 abstract Tree.Stm unCx(Label t, Label f);

 Tree.Stm unNx() {
	Label tf = new Label();
        return new Tree.SEQ(unCx(tf,tf), new Tree.LABEL(tf));
 }
}

class RelCx extends Cx {
  int relop;
  Tree.Exp left;
  Tree.Exp right;
  Tree.Stm unCx(Label t, Label f) {
     return new Tree.CJUMP(relop,left,right,t,f);
  }
  RelCx(int rel, Tree.Exp l, Tree.Exp r) {relop=rel; left=l; right=r;}
}

class SeqExp extends Exp {
   Tree.Stm a; 
   Exp b;
   SeqExp(Exp aa, Exp bb) {a=aa.unNx(); b=bb;}
   Tree.Exp unEx() {return new Tree.ESEQ(a, b.unEx());}
   Tree.Stm unNx() {return new Tree.SEQ(a, b.unNx());}
   Tree.Stm unCx(Label t, Label f) 
                    {return new Tree.SEQ(a, b.unCx(t,f));}

}

class IfThenElseExp extends Exp {
   Exp cond, a, b;
   Label t = new Label();
   Label f = new Label();
   Label join = new Label();
   IfThenElseExp(Exp cc, Exp aa, Exp bb) {cond=cc; a=aa; b=bb;}

   Tree.Stm unCx(Label tt, Label ff) {
      return new Tree.SEQ(cond.unCx(t,f),
		new Tree.SEQ(new Tree.LABEL(t),
               new Tree.SEQ(a.unCx(tt,ff),
                new Tree.SEQ(new Tree.LABEL(f),
                     b.unCx(tt,ff)))));
   }

   Tree.Exp unEx() {
        Temp r = new Temp();
	  return new Tree.ESEQ(
		  new Tree.SEQ(cond.unCx(t,f),
		   new Tree.SEQ(new Tree.LABEL(t), 
		    new Tree.SEQ(new Tree.MOVE(new Tree.TEMP(r), a.unEx()),
		     new Tree.SEQ(new Tree.JUMP(join),
		      new Tree.SEQ(new Tree.LABEL(f),
		       new Tree.SEQ(new Tree.MOVE(new Tree.TEMP(r), b.unEx()),
		        new Tree.LABEL(join))))))),
		   new Tree.TEMP(r));
   }

   Tree.Stm unNx() {
	  return new Tree.SEQ(cond.unCx(t,f),
		   new Tree.SEQ(new Tree.LABEL(t), 
		    new Tree.SEQ(a.unNx(),
		     new Tree.SEQ(new Tree.JUMP(join),
		      new Tree.SEQ(new Tree.LABEL(f),
		       new Tree.SEQ(b.unNx(),
		        new Tree.LABEL(join)))))));
   }

}

class Frags {
  private Frag first;
  void put(Frag f) {
     f.next=first;
     first=f;
  }
  Frag get() {
    Frag accum=null, f;
    while (first!=null)
     {f=first; first=f.next; f.next=accum; accum=f;}
    return accum;
  }
}

public class Translate {
  private Frame.Frame frame;
  private Frags frags;
 
  public Level outermost;

  public Translate(Frame.Frame f) {
   frame=f;
   frags=new Frags();
   outermost = new Level(f);
  }
  
  public Translate(Translate t) {
   frame=t.frame;
   frags=t.frags;
  }

  public Frag getResult() { return frags.get();}

  public Exp simpleVar(Access access, Level level, Symbol.Symbol s) {
    return new Ex(new Tree.SIMPLEVAR(s, access.fromLevel(level), access, access.home==level));
  }

  Exp bogusExp = new Ex(new Tree.CONST(0));

  Tree.Exp mem(Tree.Exp x) {return new Tree.MEM(x);}
  Tree.Exp plus(Tree.Exp a, Tree.Exp b) {
       return new Tree.BINOP(Tree.BINOP.PLUS, a, b);
  }
  boolean isconst(Tree.Exp a, int c) {
      return a instanceof Tree.CONST && ((Tree.CONST)a).value==c;
  }

  Tree.Exp mul(Tree.Exp a, Tree.Exp b) {
    if (isconst(b,4)) 
          return new Tree.BINOP(Tree.BINOP.LSHIFT,a,new Tree.CONST(2));
    else  return new Tree.BINOP(Tree.BINOP.MUL,a,b);
  }

  Tree.Stm NullStm = new Tree.EXP(new Tree.CONST(0));

  public Exp nop() {return new Nx(NullStm);}

  boolean isNop(Exp e) {
	return e instanceof Nx && isNullStm(e.unNx());
  }

  boolean isNullStm(Tree.Stm x) {
     return  x instanceof Tree.EXP 
            && ((Tree.EXP)x).exp instanceof Tree.CONST
            && ((Tree.CONST)((Tree.EXP)x).exp).value==0;
  }

  public Exp fieldVar(Exp e, int i, Symbol.Symbol f) {
    return new Ex(new 
		  Tree.FIELDVAR(e.unEx(), f, 
				 new Tree.CONST(frame.wordSize()*i)));
  }


  public Exp subscriptVar(Exp e, Exp index) {
    return new 
      Ex(new Tree.SUBSCRIPTVAR(e.unEx(), index.unEx(), 
			       new Tree.CONST(frame.wordSize())));
  }

  public Exp varExp(Exp e) { 
	return e; 
  }

  public Exp intExp(int i) { 
	return new Ex(new Tree.CONST(i)); 
  }

  public Exp stringExp(String s) {
	Label lab = new Label();
        frags.put(new DataFrag(frame.string(lab,s)));
        return new Ex(new Tree.NAME(lab));
  }

  private Tree.ExpList argExps(ExpList args) {
      if (args==null) return null;
      else return new Tree.ExpList(args.exp.unEx(), argExps(args.tail));
  }

  public Exp appExp(Label label, Level caller, Level callee, ExpList args) {
    return new Ex(new Tree.CALL(new Tree.NAME(label), 
	          new Tree.ExpList(callee.staticLink(caller),argExps(args))));

  }
  
  public Exp opExp(int oper, Exp left, Exp right) {
   Tree.Exp l=left.unEx(), r=right.unEx();
   switch (oper) {
    case Absyn.OpExp.PLUS: return new Ex(new Tree.BINOP(Tree.BINOP.PLUS,l,r));
    case Absyn.OpExp.MINUS:return new Ex(new Tree.BINOP(Tree.BINOP.MINUS,l,r));
    case Absyn.OpExp.MUL:  return new Ex(new Tree.BINOP(Tree.BINOP.MUL,l,r));
    case Absyn.OpExp.DIV:  return new Ex(new Tree.BINOP(Tree.BINOP.DIV,l,r));
    case Absyn.OpExp.EQ:   return new RelCx(Tree.CJUMP.EQ,l,r);
    case Absyn.OpExp.NE:   return new RelCx(Tree.CJUMP.NE,l,r);
    case Absyn.OpExp.LT:   return new RelCx(Tree.CJUMP.LT,l,r);
    case Absyn.OpExp.LE:   return new RelCx(Tree.CJUMP.LE,l,r);
    case Absyn.OpExp.GT:   return new RelCx(Tree.CJUMP.GT,l,r);
    case Absyn.OpExp.GE:   return new RelCx(Tree.CJUMP.GE,l,r);
   }
   throw new Error("Translate.opExp " + oper);
  }

  public Exp stringEqual(Exp a, Exp b) {
     return new Ex(frame.externalCall("stringEqual",
		new Tree.ExpList(a.unEx(), new Tree.ExpList(b.unEx(), null))));
  }

  Tree.Stm fieldList(Temp result, int i, ExpList fields) {
      if (fields==null) return NullStm;
      else return new Tree.SEQ(
	     new Tree.MOVE(mem(plus(new Tree.TEMP(result), 
				    new Tree.CONST(i))), 
		       fields.exp.unEx()),
	     fieldList(result, i + frame.wordSize(), fields.tail));
  }


  int length(ExpList x) {
   if (x==null) return 0;
   else return 1+length(x.tail);
  }

  public Exp recordExp(ExpList fields) {
     Temp result = new Temp();
     return new Ex(new Tree.ESEQ(new Tree.SEQ(
				    new Tree.MOVE(new Tree.TEMP(result), 
				    frame.externalCall("allocRecord",
					new Tree.ExpList(new Tree.CONST(
					 frame.wordSize() * length(fields)),
					 null))),
				  fieldList(result,0,fields)),
			 new Tree.TEMP(result)));
  }


  public Exp seqExp(Exp a, Exp b) {
       if (isNop(a)) return b;
       else if (isNop(b)) return a;
       else return new SeqExp(a,b);
  }
  
  public Exp assignExp(Exp var, Exp e) {
       return new Nx(new Tree.MOVE(var.unEx(), e.unEx()));
  }

  public Exp ifThenElse(Exp test, Exp thenclause, Exp elseclause) {
     return new IfThenElseExp(test,thenclause,elseclause);
  }

  public Exp ifThen(Exp test, Exp thenclause) {
     return new IfThenElseExp(test,thenclause,
			      new Nx(new Tree.EXP(new Tree.CONST(0))));
  }

  public Exp whileExp(Exp cond, Exp body, BreakScope done) {
       Label test = new Label();
       Label top = new Label();
       return new Nx(new Tree.SEQ(new Tree.JUMP(test),
		  new Tree.SEQ(new Tree.LABEL(top),
 		   new Tree.SEQ(body.unNx(),
		    new Tree.SEQ(new Tree.LABEL(test),
		     new Tree.SEQ(cond.unCx(top,done.label),
		      new Tree.LABEL(done.label)))))));
  }

  public Exp breakExp(BreakScope done) {
       return new Nx(new Tree.JUMP(done.label));
  }

  public Exp arrayExp(Exp size, Exp init) {
       return new Ex(frame.externalCall("initArray",
	      new Tree.ExpList(size.unEx(), 
	       new Tree.ExpList(init.unEx(),
		null))));
  }

  public void procEntryExit(Level level, Exp body) {
    Vector proc=level.frame.procEntryExit1(
		   new Tree.MOVE(new Tree.TEMP(level.frame.RV()),
				      body.unEx()));
    frags.put(new ProcFrag(level.frame, 
			   (Tree.Stm)(proc.elementAt(0)),
			   (Tree.Stm)(proc.elementAt(1)), 
			   (Tree.Stm)(proc.elementAt(2))));
  }
}


